@@ -49,6 +49,8 @@ from causalnex.structure.pytorch.dist_type._base import DistTypeBase
 from causalnex.structure.pytorch.nonlinear import LocallyConnected
 
 
+# Problem in pytorch 1.6 (_forward_unimplemented), fixed in next release:
+# pylint: disable=abstract-method
 class NotearsMLP(nn.Module, BaseEstimator):
     """
     Class for NOTEARS MLP (Multi-layer Perceptron) model.
@@ -157,7 +159,6 @@ class NotearsMLP(nn.Module, BaseEstimator):
         """
         return self._loc_lin_layer_weights
 
-    # pylint: disable=arguments-differ
     def forward(self, x: torch.Tensor) -> torch.Tensor:  # [n, d] -> [n, d]
         """
         Feed forward calculation for the model.
@@ -170,12 +171,14 @@ class NotearsMLP(nn.Module, BaseEstimator):
         """
         x = self.dag_layer(x)  # [n, d * m1]
         x = x.view(-1, self.dims[0], self.dims[1])  # [n, d, m1]
-        for layer in self.loc_lin_layer_weights:
+        for output_dim, layer in zip(self.dims[1:], self.loc_lin_layer_weights):
             x = torch.sigmoid(x)  # [n, d, m1]
-            # soft clamp the denominator to prevent divide by zero and prevent very large weight increases
-            x = (x - x.mean(dim=0).detach()) / torch.sqrt(
-                (self.nonlinear_clamp + x.var(dim=0).detach())
-            )
+
+            x = nn.LayerNorm(
+                output_dim,
+                eps=self.nonlinear_clamp,
+                elementwise_affine=True,
+            )(x)
 
             x = layer(x)  # [n, d, m2]
         x = x.squeeze(dim=2)  # [n, d]
@@ -193,6 +196,10 @@ class NotearsMLP(nn.Module, BaseEstimator):
             reconstructed data
         """
 
+        # perform preprocessing and column expansions, do NOT refit
+        for dist_type in self.dist_types:
+            X = dist_type.preprocess_X(X, fit_transform=False)
+
         with torch.no_grad():
             # convert the predict data to pytorch tensor
             X = torch.from_numpy(X).float().to(self.device)
@@ -252,7 +259,6 @@ class NotearsMLP(nn.Module, BaseEstimator):
             self._calculate_adj(X_torch, mean_effect=True).cpu().detach().numpy()
         )
 
-    # pylint: disable=too-many-locals
     def _dual_ascent_step(
         self, X: torch.Tensor, rho: float, alpha: float, h: float, rho_max: float
     ) -> Tuple[float, float, float]:
@@ -365,8 +371,7 @@ class NotearsMLP(nn.Module, BaseEstimator):
             X_hat = self(X)
             h_val = self._h_func()
 
-            # preallocate loss tensor
-            loss = torch.tensor(0, device=X.device)  # pylint: disable=not-callable
+            loss = 0.0
             # sum the losses across all dist types
             for dist_type in self.dist_types:
                 loss = loss + dist_type.loss(X, X_hat)
@@ -389,7 +394,8 @@ class NotearsMLP(nn.Module, BaseEstimator):
         flat_params = _get_flat_params(params)
         bounds = _get_flat_bounds(params)
 
-        while rho < rho_max:
+        h_new = np.inf
+        while (rho < rho_max) and (h_new > 0.25 * h or h_new == np.inf):
             # Magic
             sol = sopt.minimize(
                 _func,
@@ -403,8 +409,6 @@ class NotearsMLP(nn.Module, BaseEstimator):
             h_new = self._h_func().item()
             if h_new > 0.25 * h:
                 rho *= 10
-            else:
-                break
         alpha += rho * h_new
         return rho, alpha, h_new
 
@@ -425,6 +429,21 @@ class NotearsMLP(nn.Module, BaseEstimator):
             dag_layer_weight * dag_layer_weight, dim=1
         ).t()  # [i, j]
 
+        # modify the h(W) matrix to deal with expanded columns
+        original_idxs = []
+        for dist_type in self.dist_types:
+            # modify the weight matrix to prevent spurious cycles with expended columns
+            square_weight_mat = dist_type.modify_h(square_weight_mat)
+            # gather the original idxs
+            original_idxs.append(dist_type.idx)
+        # original size is largest original index
+        original_size = np.max(original_idxs) + 1
+        # subselect the top LH corner of matrix which corresponds to original data
+        square_weight_mat = square_weight_mat[:original_size, :original_size]
+        # update d and d_torch to match the new matrix size
+        d = square_weight_mat.shape[0]
+        d_torch = torch.tensor(d).to(self.device)  # pylint: disable=not-callable
+
         # h = trace_expm(a) - d  # (Zheng et al. 2018)
         characteristic_poly_mat = (
             torch.eye(d).to(self.device) + square_weight_mat / d_torch
